import re

import libsprattus
from libsprattus import DomainConstructor, ParamStrategy, Expression

def _to_DomainConstructor(obj):
    """
    Tries to convert an object to a DomainConstructor instance.

    Returns the same object if it's already a DomainConstructor. If the object
    is a list, it's converted using the Product combinator. Throws a TypeError
    if the input cannot be converted.
    """
    if type(obj) == DomainConstructor:
        return obj

    return DomainConstructor.product([_to_DomainConstructor(x) for x in obj])


def _globalize_domains(domains, gdict):
    sep = re.compile('\.|/')

    # copy simple names (not containing '.' or '/')
    for key, val in domains.items():
        if not sep.findall(key):
            gdict[key] = val

    # adapt hierarchical names
    for key, val in domains.items():
        names = sep.split(key)
        if len(names) < 2: continue

        try:
            base_name, name = names
        except ValueError:
            raise ValueError('Only two-level hierarchical domain names are '
                             'supported.')

        base = gdict.setdefault(base_name, DomainFamily(base_name))
        setattr(base, name, val)


def _DomainConstructor_call(self, *args):
    def _to_ParamStrategy(obj):
        if type(obj) == ParamStrategy:
            return obj
        elif type(obj) == Expression:
            return ParamStrategy.FromExpression(obj)
        else:
            raise ValueError("Invalid parameter to a DomainConstructor")

    dc = self
    for arg in args:
        dc = dc._parameterize(_to_ParamStrategy(arg))
    return dc

DomainConstructor.__call__ = _DomainConstructor_call


def domain(function):
    name = function.func_name
    description = function.func_doc or ""
    arity = function.func_code.co_argcount
    return DomainConstructor._makeCustom(name, description, arity, function)


class DomainFamily(object):
    def __init__(self, name):
        self._name = name

    def __repr__(self):
        return '<DomainFamily %s>' % self._name


class Config(object):
    _reserved = ['_dict', '_top_level']

    def __init__(self, _dict, top_level=True):
        self._dict = _dict
        self._top_level = top_level


    def __getattr__(self, key):
        if key in self._reserved:
            return super.__getattr__(self, key)

        if hasattr(self._dict, key):
            return getattr(self._dict, key)

        if self._top_level:
            res = self._dict.setdefault(key, {})
        else:
            if key not in self._dict:
                raise AttributeError('object has no attribute \'%s\'' % key)
            else:
                res = self._dict[key]

        if type(res) == dict:
            return Config(res)
        else:
            return res


    def __setattr__(self, key, value):
        if key in self._reserved:
            return super.__setattr__(self, key, value)

        # AbstractDomain is treated specially
        if key == 'AbstractDomain':
            self._dict[key] = _to_DomainConstructor(value)
            return

        # assignment of a dict-like object merges both dicts
        if type(self._dict.get(key)) == dict and hasattr(value, 'items'):
            for k, v in value.items():
                self._dict[key][k] = v

        else:
            self._dict[key] = value


    def __repr__(self):
        return repr(self._dict)
