#include "Config.h"

#include "utils.h"
#include "Fragment.h"
#include "AbstractValue.h"
#include "DomainConstructor.h"

#include <boost/python.hpp>

#include <cassert>
#include <vector>

namespace py = boost::python;
using namespace sprattus;

namespace sprattus
{
namespace configparser
{
shared_ptr<boost::python::dict>
Config::tryLoadingFrom(const std::string& filename)
{
    namespace py = boost::python;
    shared_ptr<py::dict> result(new py::dict());

    try {
        if (!Py_IsInitialized()) {
            Py_Initialize();

            static char obj_name[] = "path";
            py::list path(py::borrowed(PySys_GetObject(obj_name)));
            path.append(SPRATTUS_SOURCE_DIR "/python");
            path.append(SPRATTUS_BUILD_DIR "/src");
            PySys_SetObject(obj_name, path.ptr());
        }

        PyObject* module = PyImport_AddModule("__main__");
        py::dict globals(py::borrowed(PyModule_GetDict(module)));

        globals["config_dict"] = *result;
        boost::python::exec("import sprattus; from sprattus import *;"
                            "config = sprattus.Config(config_dict, True)",
                            globals, globals);

        py::dict domains;
        globals["domains"] = domains;
        for (DomainConstructor dom : DomainConstructor::all()) {
            domains[dom.name()] = py::object(dom);
        }
        boost::python::exec("sprattus._globalize_domains(domains, globals())",
                            globals, globals);

        boost::python::exec_file(filename.c_str(), globals, globals);
    } catch (...) {
        if (PyErr_Occurred()) {
            if (!PyErr_ExceptionMatches(PyExc_IOError))
                PyErr_PrintEx(1);

            PyErr_Clear();
        }

        return nullptr;
    }

    return result;
}

Config::Config(llvm::StringRef filename)
{
    if (filename == "default")
        return;

    ConfigDict_ = tryLoadingFrom(filename);
    if (!ConfigDict_) {
        ConfigDict_ = tryLoadingFrom(
            (SPRATTUS_CONFIG_SEARCH_PATH + filename + ".py").str());
    }
    if (!ConfigDict_)
        panic("Cannot load Python configuration module");
}

Config::Config()
{
    if (auto filename = std::getenv(Config::ENV_VAR)) {
        *this = Config(filename);
    } else {
        ConfigDict_.reset(new boost::python::dict());
    }
}

template <typename T> T try_extract(py::object x, T default_value)
{
    if (x.is_none())
        return default_value;

    py::extract<T> extr(x);
    if (extr.check())
        return extr;
    else
        return default_value;
}

template <typename T>
T Config::get(const char* module, const char* key, T default_value) const
{
    py::object module_entry;

    try {
        module_entry = ConfigDict_->get(module);
        if (!module_entry.is_none() &&
            PyObject_HasAttrString(module_entry.ptr(), "items")) {

            py::dict module_dict(module_entry);
            return try_extract<T>(module_dict.get(key), default_value);
        }
    } catch (py::error_already_set&) {
        // TypeError can happen if module configuration is not a dictionary
        if (!PyErr_ExceptionMatches(PyExc_TypeError)) {
            PyErr_PrintEx(1);
            panic("Invalid configuration");
        }
    }

    // Entries in the top-level dictionary that are not dictionaries themselves
    // are intepreted as an entry with a single key "Variant".
    if (std::string(key) == "Variant" && !module_entry.is_none())
        return try_extract<T>(module_entry, default_value);
    else
        return default_value;
}

template <>
DomainConstructor Config::get(const char* module, const char* key,
                              DomainConstructor default_value) const
{
    // AbstractDomain entry is special: it should be a single DomainConstructor
    // object constructed using Python combinators. It's mandatory, the
    // default_value is never used here.
    assert(std::string(module) == "AbstractDomain" &&
           std::string(key) == "Variant");

    try {
        auto ad = try_extract<DomainConstructor>(ConfigDict_->get(module),
                                                 DomainConstructor());
        if (!ad.isInvalid())
            return ad;
    } catch (py::error_already_set&) {
    }

    panic("Invalid abstract domain specifier.");
}

template int Config::get(const char* module, const char* key,
                         int default_value) const;

template bool Config::get(const char* module, const char* key,
                          bool default_value) const;

template std::string Config::get(const char* module, const char* key,
                                 std::string default_value) const;
} // namespace configparser
} // namespace sprattus
