#include "Runtime.h"

#include "ConcreteState.h"
#include "AbstractValue.h"
#include "ModuleContext.h"
#include "repr.h"

#include <cstdarg>
#include <iostream>

#include <llvm/IR/Module.h>

using namespace sprattus;

// These variables are normally declared in the instrumented module. Weak
// linkage ensures that below definitions can be discarded if another one is
// found during link-time.
extern const char
    __attribute__((weak)) sprattus_rt_database_path[] = "sprattus.db";
extern const char
    __attribute__((weak)) sprattus_rt_bitcode_path[] = "sprattus.bc";
extern const char __attribute__((weak)) sprattus_rt_config[] = "default";

namespace sprattus
{
void Runtime::evict(struct CacheEntry* ptr)
{
    if (ptr->value != nullptr) {
        ResultStore_.put(ResultStore::Key(ptr->location_id), *ptr->value);
        ptr->value.reset();
    }
}

Runtime::Runtime()
    : GlobalConfig_(sprattus_rt_config), ResultStore_(sprattus_rt_database_path)
{
    Module_ = loadModule(sprattus_rt_bitcode_path);
    assert(Module_);
    Mctx_.reset(new ModuleContext(Module_.get(), GlobalConfig_));
    int max_vals = 0;

    for (auto& func : *Module_) {
        FunctionContextV_.emplace_back(&func, Mctx_.get());
        int val_count = FunctionContextV_.back().representedValues().size();
        if (val_count > max_vals)
            max_vals = val_count;
    }

    ConcreteValues_.resize(max_vals);
}

void Runtime::trace(uint32_t function_id, uint32_t location_id,
                    const ConcreteState& cstate)
{
    ResultStore::Key key(location_id);
    struct CacheEntry* ptr = &Cache_[location_id % CACHE_SIZE];

    if (ptr->value != nullptr && ptr->location_id != location_id)
        evict(ptr);

    if (ptr->value == nullptr) {
        ptr->value = ResultStore_.get(key, FunctionContextV_[function_id]);
        ptr->location_id = location_id;
    }

    ptr->value->updateWith(cstate);
}

Runtime::~Runtime()
{
    for (unsigned i = 0; i < CACHE_SIZE; ++i)
        evict(&Cache_[i]);
}
} // namespace sprattus

extern "C" void sprattus_trace(uint32_t function_id, uint32_t location_id, ...)
{
    auto* runtime = Runtime::get();
    auto& fctx = runtime->contextFor(function_id);
    ConcreteState::Value* values = runtime->valueCache();

    // process the arguments and populate `values`
    va_list argp;
    va_start(argp, location_id);
    for (;;) {
        int bits = va_arg(argp, uint64_t);
        if (bits == 0)
            break;

#ifdef DEBUG_REPRESENTED_VALUES
        // Expensive checks but sometimes necessary for debugging. Ids of
        // represented values depend on configuration and can be fragile.
        //
        // Note that enabling this check will require recompiling sprinstrument
        // and re-instrumenting the code.
        const char* var_name = va_arg(argp, const char*);
        uint32_t rvalue_id = va_arg(argp, uint32_t);
        auto* func = fctx.getFunction();
        auto* var = func->getValueSymbolTable().lookup(var_name);
        auto* rvalue = fctx.findRepresentedValue(var);
        if (rvalue->id() != rvalue_id) {
            std::cerr << "Found mismatched RepresentedValue::id() in function "
                      << repr(func) << ":" << endl;
            std::cerr << "Variable " << var_name << " has id=" << rvalue->id()
                      << " but instrumented code says id=" << rvalue_id << endl;
            std::cerr << "All represented values: " << endl
                      << repr(fctx.representedValues()) << endl;
            abort();
        }
#else
        uint32_t rvalue_id = va_arg(argp, uint32_t);
#endif

        auto value = va_arg(argp, uint64_t);
        values[rvalue_id] = ConcreteState::Value(fctx, value, bits);
    }
    va_end(argp);

    ConcreteState cstate(fctx, values);
    runtime->trace(function_id, location_id, cstate);
}

extern "C" void sprattus_probe(int ignored, ...) {}
