#pragma once

#include "utils.h"
#include "AbstractValue.h"
#include "Expression.h"
#include "ResultStore.h"
#include "repr.h"

namespace sprattus
{
namespace domains
{
template <typename T> class Wrapper : public AbstractValue
{
  private:
    unique_ptr<AbstractValue> Value_;

  public:
    Wrapper(unique_ptr<AbstractValue>&& avalue) : Value_(std::move(avalue))
    {
        assert(dynamic_cast<T*>(Value_.get()) != nullptr);
    }

    T& value() { return *static_cast<T*>(Value_.get()); }
    const T& value() const { return *static_cast<const T*>(Value_.get()); }

    virtual void prettyPrint(PrettyPrinter& out) const
    {
        Value_->prettyPrint(out);
    }

    virtual bool joinWith(const AbstractValue& other)
    {
        auto* spec = dynamic_cast<const Wrapper<T>*>(&other);
        assert(spec);
        return Value_->joinWith(*spec->Value_.get());
    }

    virtual bool meetWith(const AbstractValue& other)
    {
        auto* spec = dynamic_cast<const Wrapper<T>*>(&other);
        assert(spec);
        return Value_->meetWith(*spec->Value_.get());
    }

    virtual bool updateWith(const ConcreteState& cstate)
    {
        return Value_->updateWith(cstate);
    }

    virtual z3::expr toFormula(const ValueMapping& vmap, z3::context& ctx) const
    {
        return Value_->toFormula(vmap, ctx);
    }

    virtual void havoc() { Value_->havoc(); }
    virtual void resetToBottom() { Value_->resetToBottom(); }
    virtual bool isTop() const { return Value_->isTop(); }
    virtual bool isBottom() const { return Value_->isBottom(); }

    virtual void widen() override { Value_->widen(); }

    virtual AbstractValue* clone() const
    {
        return new Wrapper(unique_ptr<AbstractValue>(Value_->clone()));
    }

    virtual bool isJoinableWith(const AbstractValue& av_other) const
    {
        auto* other = dynamic_cast<const Wrapper<T>*>(&av_other);
        if (other == nullptr)
            return false;
        return Value_->isJoinableWith(*other->Value_);
    }

    virtual void gatherFlattenedSubcomponents(
        std::vector<const AbstractValue*>* result) const
    {
        Value_->gatherFlattenedSubcomponents(result);
    }

    virtual void abstractConsequence(const AbstractValue& other)
    {
        Value_->abstractConsequence(other);
    }

    virtual bool operator<=(const AbstractValue& av_other) const
    {
        auto* other = dynamic_cast<const Wrapper<T>*>(&av_other);
        assert(other != nullptr);
        return *Value_ <= *other->Value_;
    }

    virtual bool operator==(const AbstractValue& av_other) const
    {
        auto* other = dynamic_cast<const Wrapper<T>*>(&av_other);
        if (other == nullptr)
            return false;

        return *Value_ == *(other->Value_);
    }
};

template <typename This, typename Base> class Cut : public Wrapper<Base>
{
    // Implemented in This:
    //     bool isInAllowedState()

  protected:
    Cut(unique_ptr<AbstractValue>&& avalue) : Wrapper<Base>(std::move(avalue))
    {
    }

  public:
    virtual bool joinWith(const AbstractValue& other)
    {
        bool changed = Wrapper<Base>::joinWith(other);
        if (changed && !static_cast<This*>(this)->isInAllowedState())
            static_cast<AbstractValue*>(this)->havoc();
        return changed;
    }

    virtual bool meetWith(const AbstractValue& other)
    {
        bool changed = Wrapper<Base>::meetWith(other);
        if (changed && !static_cast<This*>(this)->isInAllowedState())
            static_cast<AbstractValue*>(this)->resetToBottom();
        return changed;
    }

    virtual bool updateWith(const ConcreteState& cstate)
    {
        bool changed = Wrapper<Base>::updateWith(cstate);
        if (changed && !static_cast<This*>(this)->isInAllowedState())
            static_cast<AbstractValue*>(this)->havoc();
        return changed;
    }
};

class If : public AbstractValue
{
  private:
    Expression Condition_;
    unique_ptr<AbstractValue> Value_;

  public:
    If(const Expression& cond, unique_ptr<AbstractValue>&& value)
        : Condition_(cond), Value_(std::move(value))
    {
    }

    virtual void prettyPrint(PrettyPrinter& out) const
    {
        out << Condition_ << " ==> ";
        Value_->prettyPrint(out);
    }

    virtual bool joinWith(const AbstractValue& av_other)
    {
        assert(isJoinableWith(av_other));
        const If& other = static_cast<const If&>(av_other);
        return Value_->joinWith(*other.Value_);
    }

    virtual bool meetWith(const AbstractValue& other)
    {
        assert(false && "not implemented");
    }

    virtual bool updateWith(const ConcreteState& cstate)
    {
        if ((uint64_t)Condition_.eval(cstate) != 0)
            return Value_->updateWith(cstate);
        else
            return false;
    }

    virtual z3::expr toFormula(const ValueMapping& vmap, z3::context& ctx) const
    {
        return !Condition_.toFormula(vmap) || Value_->toFormula(vmap, ctx);
    }

    virtual void havoc() { Value_->havoc(); }
    virtual void resetToBottom() { Value_->resetToBottom(); }
    virtual bool isTop() const { return Value_->isTop(); }
    virtual bool isBottom() const { return Value_->isBottom(); }

    virtual bool isJoinableWith(const AbstractValue& av_other) const
    {
        auto* other = dynamic_cast<decltype(this)>(&av_other);
        if (other == nullptr)
            return false;

        return repr(Condition_) == repr(other->Condition_) &&
               Value_->isJoinableWith(*other->Value_);
    }

    virtual void widen() override { Value_->widen(); }

    virtual AbstractValue* clone() const
    {
        return new If(Condition_, unique_ptr<AbstractValue>(Value_->clone()));
    }

#ifdef ENABLE_DYNAMIC
    template <class Archive> void save(Archive& archive) const
    {
        archive(Condition_);
        archive(Value_);
    }

    template <class Archive> void load(Archive& archive) {}

    template <class Archive>
    static void load_and_construct(Archive& archive,
                                   cereal::construct<If>& construct)
    {
        Expression cond = Expression::LoadFrom(archive);
        unique_ptr<AbstractValue> val;
        archive(val);
        construct(cond, std::move(val));
    }
#endif
};
} // namespace domains
} // namespace sprattus

#ifdef ENABLE_DYNAMIC
CEREAL_REGISTER_TYPE(sprattus::domains::If);
#endif
