#!/usr/bin/env python

from __future__ import print_function

from collections import defaultdict
import sys
import json

import pygments
from pygments.formatters import HtmlFormatter
from pygments.lexers import guess_lexer, CLexer

HTML_TEMPLATE = '''
<!doctype html>
<html lang="en">
<head>
<meta charset="utf-8">
<!--[if lt IE 9]>
<script src="http://html5shiv.googlecode.com/svn/trunk/html5.js"></script>
<![endif]-->
<script type="text/javascript" src="https://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_SVG"></script>
<script src="https://code.jquery.com/jquery-2.1.4.min.js"></script>
<script>
$(document).ready(function() {
    // find out abstract values with variables with unknown source names
    var avalues_ironly = $(".source_name").filter(function(i,x) { return x.textContent == "" }).closest(".abstract_value")
    avalues_ironly.hide()

    $(".annotated").click(function() {
        $(this).next(".annotation").toggle()
    })

    $("#open_all").click(function() { $(".annotation").show() })
    $("#close_all").click(function() { $(".annotation").hide() })

    $("#llvmir_enable").change(function() {
        if ($(this).prop("checked")) {
            $(".source_name").hide();
            $(".llvmir_name").show();
            avalues_ironly.show()
        } else {
            $(".source_name").show();
            $(".llvmir_name").hide();
            avalues_ironly.hide()
        }
    })

    // assign colors to entries based on their abstract domain
    var palette = ["#DECC8C", "#DE986D", "#AB6890", "#68AB9F", "#DEC371"]
    var domains = {}
    $(".abstract_value").each(function(i, x) { domains[x.title] = ""; })
    var i = 0
    $.each(domains, function(k, v) {
        domains[k] = palette[i]
        i = i + 1
        if (i > palette.length) { i = 0 }
    })
    $(".abstract_value").each(function(i, x) {
        $(x).css({"border-left-color": domains[x.title]})
    })
})
</script>
<style>
%(pygments_css)s
.code {
    font-family: monospace;
    font-size: 14px;
    width: 60em;
    display: block;
    margin-left: auto;
    margin-right: auto;
    margin-top: 3em;
    border: 1px solid #ccc;
    border-radius: 3px;
    padding-top: 0.3em;
    padding-bottom: 0.3em;
}

#controls {
    position: fixed;
    left: 0;
    top: 0;
    font-size: 14px;
    font-family: sans-serif;
    width: 100%%;
    padding: 0.5em;
    margin: 0;
    border-bottom: 1px solid #ccc;
    background: #f0f0f0;
    vertical-align: center;
}

.codeline { white-space: pre-wrap }

.lineno {
    width: 3em;
    display: inline-block;
    text-align: right;
    padding-right: 0.5em;
    font-weight: bold;
}

.annotated {
    cursor: pointer;
}

.annotated:hover {
    background-color: #DFDFDF;
}

.annotated .lineno {
    color: #CC0000;
}

.column_padding {
    font-family: monospace;
    white-space: pre;
}

.annotation {
    font-family: serif;
    padding-top: 0.5ex;
    padding-bottom: 0.5ex;
    padding-left: 4.5ex;
    background-color: white;
    text-align: left;
    display: none;
}

.source_name {
    font-family: monospace;
}

.llvmir_name {
    font-family: monospace;
    display: none;
}

.abstract_value {
    padding-left: 0.5ex;
    border-left: 1ex solid blue;
    margin-bottom: 1px;
}
</style>
</head>
<body class="tex2jax_ignore">
<div id="controls">
<button type="button" id="open_all">Open All</button>
<button type="button" id="close_all">Close All</button>
<input type="checkbox" id="llvmir_enable"><label for="llvmir_enable">Display LLVM IR names</label></input>
</div>
%(contents)s
</body>
</html>
'''

class Formatter(HtmlFormatter):
    def __init__(self, annotations):
        HtmlFormatter.__init__(self, style='friendly')
        self.annotations = annotations

    def wrap(self, source, outfile):
        lineno = 0
        yield 0, '<div class="code">'
        for i, txt in source:
            if i == 1:
                lineno = lineno + 1
                orig_line = txt
                txt = txt.replace('\t', ' ' * 8)
                txt = '<span class="lineno">' + str(lineno) + '.</span>' + txt

                if len(self.annotations[lineno]) > 0:
                    txt = '<div class="annotated codeline">' + txt + '</div>'
                    for ann_txt, column in self.annotations[lineno]:
                        txt += '<div class="annotation tex2jax_process">'

                        if column:
                            txt += '<span class="column_padding">'
                            for x in (orig_line + column * ' ')[:column-1]:
                                if x == '\t':
                                    txt += ' ' * 8
                                else:
                                    txt += ' '
                            txt += '</span>&#11025;&nbsp;' + ann_txt
                        else:
                            txt += ann_txt

                        txt += '</div>'
                else:
                    txt = '<div class="codeline">' + txt + '</div>'

            yield i, txt
        yield 0, '</div>'


def main(in_fp, out_fp=sys.stdout):
    input_spec = json.load(in_fp)

    with open(input_spec["source"]) as fp:
        code = fp.read()

    annotations = defaultdict(lambda: [])
    for entry in input_spec["annotations"]:
        if entry.get('column'):
            annotations[entry['line']].append((entry['annotation'], entry['column']))
        else:
            annotations[entry['line']].append((entry['annotation'], None))

    fmter = Formatter(annotations)
    lexer = CLexer()
    try:
        lexer = guess_lexer(code)
    except: pass

    out_fp.write(HTML_TEMPLATE % {
            'pygments_css': fmter.get_style_defs('.code'),
            'contents': pygments.highlight(code, lexer, fmter),
        })


if __name__ == '__main__':
    if len(sys.argv) > 1:
        main(open(sys.argv[1]))
    else:
        main(sys.stdin)
