/**
 * This simple command line tool is intended to help to verify the outputs in
 * automated tests.
 *
 * It reads a bitcode file and finds calls to a function sprattus_probe:
 *
 *     void sprattus_probe(int ignore, ...);
 *
 * The first argument doesn't matter but has to be there since C variadic
 * functions require that. LLVM IR names of remaining arguments are determined
 * and the tool replaces each occurrence of any of these names in the standard
 * input with the string "probeN" where N is the index of the argument.
 *
 * The input is naively tokenized so the substitution might not work if the
 * name is in parenthesis or quotation marks.
 *
 * This tool is used in some C test cases where the LLVM IR names of
 * interesting variables are not known a priori but are used in the output of
 * some tools like `spranalyze`.
 */

#include "utils.h"
#include "repr.h"

#include <iostream>
#include <map>

#include <llvm/IR/Module.h>
#include <llvm/IR/Instructions.h>
#include <llvm/Support/CommandLine.h>
#include <llvm/Bitcode/ReaderWriter.h>

using namespace llvm;

cl::opt<std::string> BitcodeFile(cl::Positional, cl::desc("<bitcode.bc>"),
                                 cl::Required);

int main(int argc, char** argv)
{
    cl::ParseCommandLineOptions(argc, argv);
    auto module = sprattus::loadModule(BitcodeFile);
    Function* probe_func = module->getFunction("sprattus_probe");

    // prepare a substitution list
    std::map<std::string, std::string> substitutions;
    if (probe_func) {
        for (auto& use : probe_func->uses()) {
            auto* as_call = dyn_cast<CallInst>(use.getUser());
            if (!as_call)
                continue;

            // skip the first argument, it has to be there since C doesn't
            // support f(...) functions
            for (unsigned i = 1; i < as_call->getNumOperands() - 1; ++i) {
                Value* val = as_call->getOperand(i);
                std::string name = val->getName();
                if (name.length() > 0) {
                    assert(substitutions.find(name) == substitutions.end());
                    substitutions[name] = "probe" + std::to_string(i);
                }
            }
        }
    }

    // apply the substitutions to the standard input line-by-line
    std::string line;
    const char* DELIM = " \t";
    while (std::getline(std::cin, line)) {
        if (line.size() == 0) {
            std::cout << std::endl;
            continue;
        }

        char* c_line = &line[0]; // allowed in C++11 for non-empty strings
        char* tok = std::strtok(c_line, DELIM);

        for (; tok != nullptr; tok = std::strtok(nullptr, DELIM)) {
            auto itr = substitutions.find(tok);

            if (itr == substitutions.end())
                std::cout << tok;
            else
                std::cout << itr->second;

            std::cout << " ";
        }

        std::cout << std::endl;
    }

    return 0;
}
