#include "utils.h"

#include "ModuleContext.h"
#include "FunctionContext.h"
#include "Analyzer.h"
#include "MemoryModel.h"
#include "Config.h"
#include "ResultStore.h"
#include "repr.h"
#include "domains/MemRegions.h"

#include <llvm/Support/CommandLine.h>
#include <llvm/IR/Module.h>
#include <llvm/Bitcode/ReaderWriter.h>
#include <llvm/Support/MemoryBuffer.h>

#include <csignal>
#include <iomanip>

using namespace sprattus;

namespace // unnamed
{
void printVersion() { std::cout << "Sprattus " << SPRATTUS_VERSION << endl; }

void onAbort(int)
{
    std::cerr << "......_____........................................" << endl
              << "     /     \\/|                                   " << endl
              << "     \\o__  /\\|                                  " << endl
              << "         \\|                                      " << endl;
}

void listDomains()
{
    std::cout << "# All available abstract domains: " << endl;

    int width = 0;
    for (auto& domain : DomainConstructor::all()) {
        int len = domain.name().length();
        if (len > width)
            width = len;
    }
    width++;

    for (auto& domain : DomainConstructor::all()) {
        std::cout << "    - ";
        std::cout << std::setw(width) << std::left << domain.name();
        std::cout << "# " << domain.description() << endl;
    }
}
} // namespace unnamed

namespace cl = llvm::cl;

enum where_t { EXIT, EVERYWHERE, MEMORY };

cl::opt<std::string> InputFilename(cl::Positional, cl::desc("<file.bc>"),
                                   cl::init("file.bc"));

cl::opt<std::string> FunctionName(
    "function",
    cl::desc("Specify the function to be analyzed (default: first of module)"),
    cl::value_desc("name"), cl::init("-"));

cl::opt<std::string> ConfigFilename(
    "config", cl::desc("Specify the configuration file to use for analysis"),
    cl::value_desc("cfg.py"), cl::init("default"));

cl::opt<bool>
    CheckAssertions("check-assertions",
                    cl::desc("Check for possibly violated assertions"));

cl::opt<bool>
    CheckMemSafety("check-memsafety",
                   cl::desc("Check for possibly invalid memory accesses"));

cl::opt<bool> NotInSSAForm(
    "not-in-ssa-form",
    cl::desc(
        "Allows to analyze code that hasn't been preprocessed with `mem2reg`"));

cl::opt<bool, true> Verbose("verbose",
                            cl::desc("Print additional informational messages"),
                            cl::location(VerboseEnable));

cl::alias VerboseA("v", cl::desc("Alias for -verbose"), cl::aliasopt(Verbose));

cl::opt<Analyzer::mode_t, Analyzer::FULL> AnalyzerMode(
    "mode", cl::desc("Analyzer mode"),
    cl::values(
        clEnumValN(Analyzer::FULL, "full", "Full, sound, static analysis"),
        clEnumValN(Analyzer::ONLY_DYNAMIC, "only-dynamic",
                   "Just print the dynamic results from the input database"),
        clEnumValN(Analyzer::ABS_POINTS_DYNAMIC, "abs-points-dynamic",
                   "An unsound mode in which results in abstraction points are "
                   "taken from the input database but the rest is computed "
                   "statically."),
        clEnumValEnd));

cl::opt<where_t, EXIT> Where(
    "where", cl::desc("At which program points should the results be computed"),
    cl::values(
        clEnumValN(EXIT, "exit", "at the EXIT node"),
        clEnumValN(EVERYWHERE, "everywhere", "after every basic block"),
        clEnumValN(MEMORY, "memory",
                   "after every basic block that contains memory operations"),
        clEnumValEnd));

cl::opt<bool, false>
    OutputJSON("json",
               cl::desc("Produce output in a machine-processable JSON format"));

cl::opt<bool, false> ListDomains(
    "list-domains",
    cl::desc(
        "Lists all available abstract domains along with their documentation"));

namespace // unnamed
{
bool wantsResultsFor(llvm::BasicBlock* bb)
{
    using namespace llvm;

    switch (Where) {
    case EXIT:
        return bb == nullptr;

    case EVERYWHERE:
        return true;

    case MEMORY:
        if (bb == nullptr)
            return false;

        for (auto& inst : *bb) {
            if (isa<LoadInst>(inst) || isa<StoreInst>(inst))
                return true;
        }
        return false;

    default:
        llvm_unreachable("bad `-where=` value");
    }
}
}

int main(int argc, char** argv)
{
    using namespace sprattus;

    std::signal(SIGABRT, onAbort);
    llvm::cl::SetVersionPrinter((*printVersion));
    llvm::cl::ParseCommandLineOptions(argc, argv);

    if (ListDomains) {
        listDomains();
        return 0;
    }

    configparser::Config cfg(ConfigFilename);
    unique_ptr<llvm::Module> module = loadModule(InputFilename);
    if (module == nullptr) {
        panic("Failed to read input file.");
    }

    auto config_val =
        ModuleContext::readGlobalString(module.get(), "sprattus_rt_config");

    if (config_val.size() > 0 && ConfigFilename == "default") {
        cfg = configparser::Config(config_val);
        vout << "Using configuration `" << config_val
             << "` stored in the bitcode file by `sprinstrument`" << endl;
    }

    auto mctx = make_unique<const ModuleContext>(module.get(), cfg);

    llvm::Function* function = NULL;
    for (auto& f : *module) {
        if (f.getBasicBlockList().empty())
            continue;
        if (FunctionName == "-") {
            // no function specified -> take the first
            function = &f;
            break;
        }
        if (f.getName().str() == FunctionName) {
            function = &f;
            break;
        }
    }
    if (function == NULL) {
        panic("Function not found.");
    }

    if (!NotInSSAForm && !isInSSAForm(function))
        panic("Input function not in SSA form. Preprocess with `-mem2reg`.");

    auto fctx = mctx->createFunctionContext(function);
    auto fragment_decomp = FragmentDecomposition::For(*fctx.get());
    vout << "Fragment decomposition: " << fragment_decomp << endl;
    DomainConstructor domain(fctx->getConfig());

    auto analyzer =
        Analyzer::New(*fctx.get(), fragment_decomp, domain, AnalyzerMode);

    if (CheckAssertions) {
        int num_violations = 0;
        for (auto& bb : *function) {
            for (auto& instr : bb) {
                if (llvm::isa<llvm::CallInst>(instr)) {
                    auto& call = llvm::cast<llvm::CallInst>(instr);
                    if (call.getCalledValue()->getName().str() ==
                        "__assert_fail") {

                        if (!analyzer->at(&bb)->isBottom()) {
                            num_violations++;
                            std::cout << std::endl
                                      << "Violated assertion at "
                                      << bb.getName().str()
                                      << ". Computed result:" << std::endl
                                      << *analyzer->at(&bb);
                        }
                    }
                }
            }
        }
        if (num_violations) {
            std::cout << "===================================================="
                      << "====================" << std::endl
                      << "  " << num_violations << " violated assertion"
                      << ((num_violations == 1) ? "" : "s") << " detected.";
        } else {
            std::cout << "No violated assertions detected.";
        }

        std::cout << std::endl;
        return (num_violations < 128) ? num_violations : 1;
    }

    if (CheckMemSafety) {
        int num_violations = 0;
        for (auto& bb : *function) {
            bool contains_mem_op = false;
            for (auto& inst : bb) {
                if (llvm::isa<llvm::StoreInst>(inst) ||
                    llvm::isa<llvm::LoadInst>(inst))
                    contains_mem_op = true;
            }
            if (!contains_mem_op)
                continue;
            std::vector<const AbstractValue*> vals;
            analyzer->after(&bb)->gatherFlattenedSubcomponents(&vals);
            for (auto& instr : bb) {
                llvm::Value* ptr = nullptr;
                if (auto as_store = llvm::dyn_cast<llvm::StoreInst>(&instr)) {
                    ptr = as_store->getPointerOperand();
                }
                if (auto as_load = llvm::dyn_cast<llvm::LoadInst>(&instr)) {
                    ptr = as_load->getPointerOperand();
                }
                if (ptr) {
                    bool is_okay = false;
                    for (auto v : vals) {
                        if (auto as_vr =
                                dynamic_cast<const domains::ValidRegion*>(v)) {
                            if (as_vr->getRepresentedPointer() == ptr &&
                                as_vr->isValid()) {
                                is_okay = true;
                                break;
                            }
                        }
                    }
                    if (!is_okay) {
                        num_violations++;
                        std::cout << "Possibly invalid memory access to "
                                  << repr(ptr) << " at " << bb.getName().str()
                                  << std::endl;
                    } else {
                        std::cout << "Definitely valid memory access to "
                                  << repr(ptr) << " at " << bb.getName().str()
                                  << std::endl;
                    }
                }
            }
        }
        if (num_violations) {
            std::cout << std::endl
                      << "===================================================="
                      << "====================" << std::endl
                      << " " << num_violations
                      << " possibly invalid memory access"
                      << ((num_violations == 1) ? "" : "es") << " detected.";
        } else {
            std::cout << std::endl
                      << "===================================================="
                      << "====================" << std::endl
                      << "No possibly invalid memory accesses detected.";
        }

        std::cout << std::endl;
        return (num_violations < 128) ? num_violations : 1;
    }

    if (OutputJSON) {
        JsonAnnotationOutput json_out(&std::cout, function);

        // prepare the list of program points we want to analyze
        std::set<llvm::BasicBlock*> abs_points;
        if (Where == EVERYWHERE) {
            for (llvm::BasicBlock& bb : *function)
                abs_points.insert(&bb);
        } else if (Where == EXIT) {
            for (auto& frag : fragment_decomp) {
                abs_points.insert(frag.getStart());
                abs_points.insert(frag.getEnd());
            }
        } else {
            std::cerr << "JSON output not implemented for this -where= value"
                      << endl;
            abort();
        }

        // Fragment::EXIT has to be handled specially
        auto itr = abs_points.find(Fragment::EXIT);
        if (itr != abs_points.end())
            abs_points.erase(itr);

        PrettyPrinter exit_res_out(true);
        analyzer->at(Fragment::EXIT)->prettyPrint(exit_res_out);

        for (auto& bb : *function) {
            if (llvm::succ_begin(&bb) == llvm::succ_end(&bb)) {
                // make sure we don't emit duplicate entries below
                abs_points.erase(abs_points.find(&bb));

                int line = bb.getTerminator()->getDebugLoc().getLine();
                json_out.emit(exit_res_out.str(), line);
            }
        }

        for (auto* location : abs_points) {
            int line = -1;
            for (auto& instr : *location) {
                if (!llvm::isa<llvm::PHINode>(instr)) {
                    line = instr.getDebugLoc().getLine();
                    break;
                }
            }

            json_out.emit(analyzer->at(location), line);
        }
    } else {
        if (Where == EXIT) {
            std::cout << *analyzer->at(Fragment::EXIT) << endl;
        } else {
            for (llvm::BasicBlock& bb : *function) {
                if (!wantsResultsFor(&bb))
                    continue;

                std::cout << repr(bb) << ":" << endl;

                // Most testcases want the `Analyzer::after()` result here but
                // ones from test/dynamic/ that don't do the static analysis
                // cneed results from `Analyzer::at()` (since these are the
                // only ones that can be provided by the dynamic analysis).
                // SOLUTION: Use at() if after() gives bottom
                auto* result = analyzer->after(&bb);
                if (result->isBottom())
                    result = analyzer->at(&bb);

                std::cout << *result << endl;
                std::cout << "============================================"
                          << endl;
            }
            if (wantsResultsFor(nullptr)) {
                std::cout << "EXIT:" << endl
                          << *analyzer->at(Fragment::EXIT) << endl;
            }
        }
    }

    return 0;
}
