#include "utils.h"

#include "DomainConstructor.h"
#include "FragmentDecomposition.h"
#include "ModuleContext.h"
#include "ResultStore.h"
#include "Config.h"
#include "Runtime.h"
#include "RepresentedValue.h"

#include <cstdlib>
#include <fstream>

#include <llvm/IR/Instructions.h>
#include <llvm/IR/BasicBlock.h>
#include <llvm/IR/DerivedTypes.h>
#include <llvm/IR/Constants.h>
#include <llvm/Support/CommandLine.h>
#include <llvm/Support/raw_os_ostream.h>
#include <llvm/Support/FileSystem.h>
#include <llvm/Bitcode/ReaderWriter.h>

using namespace sprattus;

namespace cl = llvm::cl;

cl::opt<std::string> InputFilename(cl::Positional, cl::desc("<input.bc>"),
                                   cl::Required);

cl::opt<std::string> OutputDBFilename(cl::Positional, cl::desc("<output.db>"),
                                      cl::Required);

cl::opt<std::string> OutputBCFilename(cl::Positional, cl::desc("<output.bc>"),
                                      cl::Optional, cl::init(""));

cl::opt<std::string>
    ConfigName("config",
               cl::desc("The configuration file to use for the analysis"),
               cl::value_desc("config.py"), cl::init("default"));

cl::opt<bool, false> InitDBOnly(
    "init-db-only",
    cl::desc("Don't instrument the code, only initialize the database"));

enum which_t { ALL, MEM };

cl::opt<which_t, ALL> WhichFunctions(
    "which", cl::desc("Which functions to instrument"),
    cl::values(
        clEnumValN(ALL, "all", "instrument all the functions"),
        clEnumValN(MEM, "memory",
                   "only instrument functions that contain memory accesses"),
        clEnumValEnd));

std::string getRealpath(const std::string& path)
{
    llvm::SmallVector<char, 64> as_vect(path.c_str(),
                                        path.c_str() + path.length());

    std::error_code err = llvm::sys::fs::make_absolute(as_vect);
    if (err) {
        std::cerr << "Invalid path: " << path << endl;
        abort();
    }

    return std::string(as_vect.begin(), as_vect.end());
}

llvm::Value* getStringPointer(llvm::Module* module, std::string name,
                              std::string value)
{
    using namespace llvm;
    auto& ctx = module->getContext();

    Constant* init = ConstantDataArray::getString(ctx, value);
    GlobalVariable* global = dyn_cast<GlobalVariable>(
        module->getOrInsertGlobal(name, init->getType()));
    assert(global != nullptr);
    global->setInitializer(init);
    return global;
}

uint32_t functionId(llvm::Function* func)
{
    using namespace llvm;
    Module* module = func->getParent();
    uint32_t result = 0;
    for (auto& f : *module) {
        if (&f == func)
            break;
        result++;
    }
    return result;
}

llvm::CallInst* insertTraceCall(const FunctionContext& fctx,
                                llvm::BasicBlock* location,
                                llvm::Instruction* before, bool after = false)
{
    using namespace llvm;

    // find the called function `sprattus_trace`
    Module* module = fctx.getFunction()->getParent();
    Type* i32ty = Type::getInt32Ty(module->getContext());
    Type* i64ty = Type::getInt64Ty(module->getContext());
    auto* ftype = FunctionType::get(Type::getVoidTy(module->getContext()),
                                    ArrayRef<Type*>({i32ty, i32ty}), true);
    auto* trace_func = module->getOrInsertFunction("sprattus_trace", ftype);

    // construct the call
    Function* func = fctx.getFunction();
    ResultStore::Key key(*func, location);
    std::vector<Value*> args = {ConstantInt::get(i32ty, functionId(func)),
                                ConstantInt::get(i32ty, key.getId())};
    for (Value* value : fctx.valuesAvailableIn(location, after)) {
        Type* type = value->getType();

        if (!type->isIntegerTy() && !type->isPointerTy())
            continue;

        unsigned int bits = 0;

        if (type->isPointerTy()) {
            bits = fctx.getPointerSize();
        } else {
            z3::sort sort = fctx.sortForType(value->getType());
            assert(sort.is_bv() && "not implemented");
            bits = sort.bv_size();
        }

        args.push_back(ConstantInt::get(i64ty, bits));
#ifdef DEBUG_REPRESENTED_VALUES
        // necessary for the extra variable check in sprattus_trace
        std::string val_name = value->getName().str();
        args.push_back(
            getStringPointer(module, "__SPRATTUS_VAR_" + val_name, val_name));
#endif
        args.push_back(
            ConstantInt::get(i32ty, fctx.findRepresentedValue(value)->id()));

        // currently all values are passed as 64-bit integers, zero-extended
        // if necessary
        Value* dyn_value = value;
        if (type->isPointerTy())
            dyn_value = new PtrToIntInst(value, i64ty, "", before);
        if (dyn_value->getType()->getIntegerBitWidth() < 64)
            dyn_value = new ZExtInst(dyn_value, i64ty, "", before);

        assert(dyn_value->getType()->getIntegerBitWidth() == 64);
        args.push_back(dyn_value);
    }
    args.push_back(ConstantInt::get(i64ty, 0));

    auto res = CallInst::Create(trace_func, args, "", before);

    // add metadata so that we recognize these calls in the static analysis
    LLVMContext& C = res->getContext();
    MDNode* N = MDNode::get(C, MDString::get(C, "sprattus_trace"));
    res->setMetadata("sprattus", N);

    return res;
}

void instrumentBB(const FunctionContext& fctx, llvm::BasicBlock* bb)
{
    // insert an instrumentation call before the first non-phi instruction
    auto itr = bb->begin();
    while (llvm::isa<llvm::PHINode>(*itr))
        ++itr;
    insertTraceCall(fctx, bb, &*itr);
}

/**
 * Generate instrumentation code for the Fragment::EXIT location. This can
 * result in multiple class to sprattus_trace if there are multiple return
 * instructions in the function.
 */
void instrumentEXIT(const FunctionContext& fctx)
{
    using namespace llvm;
    auto* function = fctx.getFunction();

    for (auto& bb : *function) {
        auto* term = bb.getTerminator();
        if (isa<ReturnInst>(term))
            insertTraceCall(fctx, nullptr, bb.getTerminator(), true);
    }
}

bool shouldBeInstrumented(llvm::Function* function)
{
    using namespace llvm;

    switch (WhichFunctions) {
    case ALL:
        return true;

    case MEM:
        for (auto& bb : *function) {
            for (auto& inst : bb) {
                if (isa<LoadInst>(inst) || isa<StoreInst>(inst))
                    return true;
            }
        }
        return false;

    default:
        llvm_unreachable("invalid 'which' value");
    }
}

int main(int argc, char** argv)
{
    cl::ParseCommandLineOptions(argc, argv);
    configparser::Config config(ConfigName);
    DomainConstructor domain(config);
    auto module = loadModule(InputFilename);

    auto mctx = make_unique<const ModuleContext>(module.get(), config);

    OutputDBFilename = getRealpath(OutputDBFilename);
    OutputBCFilename = getRealpath(OutputBCFilename);

    // delete the DB file if it exists
    std::remove(OutputDBFilename.c_str());

    ResultStore store(OutputDBFilename);

    std::vector<llvm::Function*> to_instrument;
    for (auto& function : *module) {
        if (function.begin() == function.end())
            continue;

        if (shouldBeInstrumented(&function))
            to_instrument.push_back(&function);
    }

    int i = -1;
    for (auto* function : to_instrument) {
        i++;

        std::cout << "[" << i << "/" << to_instrument.size() << "]\t"
                  << function->getName().str() << endl;

        auto fctx = mctx->createFunctionContext(function);
        auto fragment_decomp = FragmentDecomposition::For(*fctx.get());

        for (auto* bb : fragment_decomp.abstractionPoints()) {
            auto avalue = domain.makeBottom(*fctx.get(), bb, false);
            ResultStore::Key key(*function, bb);
            store.put(key, *avalue);

            if (!InitDBOnly) {
                if (bb == Fragment::EXIT)
                    instrumentEXIT(*fctx.get());
                else
                    instrumentBB(*fctx.get(), bb);
            }
        }
    }

    // set up sprattus_rt_* variales
    getStringPointer(module.get(), "sprattus_rt_database_path",
                     OutputDBFilename);
    getStringPointer(module.get(), "sprattus_rt_bitcode_path",
                     OutputBCFilename);
    getStringPointer(module.get(), "sprattus_rt_config", ConfigName);

    if (OutputBCFilename.size() > 0) {
        std::ofstream out(OutputBCFilename);
        llvm::raw_os_ostream raw_out(out);
        llvm::WriteBitcodeToFile(module.get(), raw_out);
    }

    return 0;
}
