#include <memory>
#include <iterator>

#include "./common.h"

#include "ModuleContext.h"
#include "FunctionContext.h"
#include "Analyzer.h"
#include "Config.h"
#include "Fragment.h"
#include "FragmentDecomposition.h"
#include "repr.h"

using namespace sprattus;
using namespace llvm;

const char* source = "define i1 @test(i1 %x) #0 {\n"
                     "b00:\n"
                     "  br i1 %x, label %b01, label %b02\n"
                     "b01:\n"
                     "  br label %b02\n"
                     "b02:\n"
                     "  br label %b03\n"
                     "b03:\n"
                     "  br i1 %x, label %b04, label %b10\n"
                     "b04:\n"
                     "  br i1 %x, label %b05, label %b06\n"
                     "b05:\n"
                     "  br label %b08\n"
                     "b06:\n"
                     "  unreachable\n"
                     "b07:\n"
                     "  br label %b08\n"
                     "b08:\n"
                     "  br label %b09\n"
                     "b09:\n"
                     "  br label %b03\n"
                     "b10:\n"
                     "  ret i1 %x\n"
                     "}\n";

template <typename T> std::string separatedRepr(const T& xs)
{
    std::set<std::string> sorted_repr;
    for (auto& x : xs)
        sorted_repr.insert(repr(x));

    std::ostringstream out;
    bool needs_comma = false;
    for (auto& x : sorted_repr) {
        if (needs_comma)
            out << ",";

        out << x;
        needs_comma = true;
    }

    return out.str();
}

TEST(FragmentDecompositionTests, Edges)
{
    auto module = parseAssembly(source);
    auto apoints = FragmentDecomposition::GetAbstractionPoints(
        &*module->begin(), FragmentDecomposition::Edges);

    std::string correct = "b00,b01,b02,b03,b04,b05,b06,b07,b08,b09,b10,nullptr";
    EXPECT_EQ(correct, separatedRepr(apoints));
}

TEST(FragmentDecompositionTests, Headers)
{
    auto module = parseAssembly(source);
    auto apoints = FragmentDecomposition::GetAbstractionPoints(
        &*module->begin(), FragmentDecomposition::Headers);

    EXPECT_EQ("b00,b03,b07,nullptr", separatedRepr(apoints));

    auto mctx = make_unique<const ModuleContext>(module.get(),
                                                 configparser::Config());

    auto fctx = mctx->createFunctionContext(&*module->begin());
    auto decomp = FragmentDecomposition::ForAbstractionPoints(*fctx.get(),
                                                              apoints);
    for (auto& frag : decomp) {
        if (repr(frag) == "b00-->b03") {
            EXPECT_EQ("b00,b01,b02,b03", separatedRepr(frag.locations()));
            EXPECT_EQ("(b00, b01),(b00, b02),(b01, b02),(b02, b03)",
                      separatedRepr(frag.edges()));
        }
    }
}

TEST(FragmentDecompositionTests, Backedges)
{
    auto module = parseAssembly(source);
    auto apoints = FragmentDecomposition::GetAbstractionPoints(
        &*module->begin(), FragmentDecomposition::Backedges);

    EXPECT_EQ("b00,b03,b07,b09,nullptr", separatedRepr(apoints));
}

TEST(FragmentDecompositionTests, SelfLoop)
{
    const char* source = "define i1 @test(i1 %x) #0{\n"
                         "entry:\n"
                         "  br label %loop\n"
                         "loop:\n"
                         "  br label %loop\n"
                         "}\n";

    auto module = parseAssembly(source);
    auto mctx = make_unique<const ModuleContext>(module.get(),
                                                 configparser::Config());
    auto fctx = mctx->createFunctionContext(&*module->begin());

    for (auto strategy :
         {FragmentDecomposition::Edges, FragmentDecomposition::Headers,
          FragmentDecomposition::Backedges, FragmentDecomposition::Body}) {

        auto apoints = FragmentDecomposition::GetAbstractionPoints(
            &*module->begin(), strategy);

        EXPECT_EQ("entry,loop,nullptr", separatedRepr(apoints));
    }

    auto apoints = FragmentDecomposition::GetAbstractionPoints(
        &*module->begin(), FragmentDecomposition::Edges);
    auto decomp = FragmentDecomposition::ForAbstractionPoints(*fctx.get(),
                                                              apoints);

    EXPECT_EQ("entry-->loop,loop-->loop", separatedRepr(decomp));

    for (auto& frag : decomp) {
        if (repr(frag) == "entry-->loop") {
            EXPECT_EQ("entry,loop", separatedRepr(frag.locations()));
            EXPECT_EQ("(entry, loop)", separatedRepr(frag.edges()));
        } else {
            EXPECT_EQ("loop", separatedRepr(frag.locations()));
            EXPECT_EQ("(loop, loop)", separatedRepr(frag.edges()));
        }
    }
}
