#include <memory>
#include <iterator>

#include "./common.h"

#include "AbstractValue.h"
#include "FunctionContext.h"
#include "ModuleContext.h"
#include "Fragment.h"
#include "Config.h"

using namespace sprattus;
using namespace llvm;

TEST(FragmentTests, SingleEdge)
{
    const char* src = "define i32 @test(i32 %a) #0 {\n"
                      "bb:\n"
                      "   %x = add i32 %a, 1\n"
                      "   ret i32 %a\n"
                      "}\n";

    auto module = parseAssembly(src);
    Function* func = &*module->begin();
    BasicBlock* bb = &func->getEntryBlock();

    auto mctx = make_unique<const ModuleContext>(module.get(),
                                                 configparser::Config());
    auto fctx = mctx->createFunctionContext(func);
    std::vector<Fragment::edge> edges = {Fragment::edge(bb, Fragment::EXIT)};
    Fragment frag(*fctx.get(), bb, Fragment::EXIT, edges);

    EXPECT_EQ(frag.edges().size(), (size_t)1);
    EXPECT_EQ(*frag.edges().begin(), edges[0]);

    std::set<BasicBlock*> locations = {bb, Fragment::EXIT};
    EXPECT_EQ(frag.locations(), locations);

    EXPECT_EQ(frag.edgesFrom(bb).size(), (size_t)1);
    EXPECT_EQ(frag.edgesFrom(bb)[0], edges[0]);
    EXPECT_EQ(frag.edgesFrom(Fragment::EXIT).size(), (size_t)0);

    EXPECT_EQ(frag.edgesTo(Fragment::EXIT).size(), (size_t)1);
    EXPECT_EQ(frag.edgesTo(Fragment::EXIT)[0], edges[0]);
    EXPECT_EQ(frag.edgesTo(bb).size(), (size_t)0);

    Instruction* x = &*bb->begin();
    Instruction* ret = &*(++bb->begin());

    ASSERT_EQ(x->getName().str(), "x");
    EXPECT_TRUE(frag.reachable(x, ret));
    EXPECT_FALSE(frag.reachable(ret, x));
    EXPECT_TRUE(frag.reachable(x, x));
    EXPECT_TRUE(frag.reachable(ret, ret));

    EXPECT_TRUE(frag.defines(x));
    EXPECT_TRUE(frag.defines(ret));
}

TEST(FragmentTests, SelfLoop)
{
    const char* src = "define i32 @test(i32 %a) #0 {\n"
                      "bb:\n"
                      "   br label %loop\n"
                      "loop:\n"
                      "   %x = phi i32 [%a, %bb], [%x_next, %loop]\n"
                      "   %x_next = add i32 %x, 1\n"
                      "   br label %loop\n"
                      "}\n";

    auto module = parseAssembly(src);
    Function* func = &*module->begin();
    BasicBlock* loop = &*(++func->begin());
    auto insts = pointerVector(loop->getInstList());

    auto mctx = make_unique<const ModuleContext>(module.get(),
                                                 configparser::Config());
    auto fctx = mctx->createFunctionContext(func);

    std::vector<Fragment::edge> edges = {Fragment::edge(loop, loop)};
    Fragment frag(*fctx.get(), loop, loop, edges);

    EXPECT_TRUE(frag.reachable(insts[1], insts[0]));
    EXPECT_FALSE(frag.reachable(insts[0], insts[1]));
    EXPECT_TRUE(frag.reachable(insts[1], insts[2]));
}

TEST(FragmentTests, DagReachability)
{
    const char* src = "define i32 @test(i1 %a) #0 {\n"
                      "bb:\n"
                      "   br i1 %a, label %A, label %B\n"
                      "A:\n"
                      "   br label %C\n"
                      "B:\n"
                      "   br label %C\n"
                      "C:\n"
                      "   br i1 %a, label %D, label %E\n"
                      "D:\n"
                      "   br label %E\n"
                      "E:\n"
                      "   ret i32 0\n"
                      "}\n";

    auto module = parseAssembly(src);
    Function* func = &*module->begin();

    BasicBlock* bb, *A, *B, *C, *D, *E;
    unpack(std::tie(bb, A, B, C, D, E),
           pointerVector(func->getBasicBlockList()));

    auto mctx = make_unique<const ModuleContext>(module.get(),
                                                 configparser::Config());
    auto fctx = mctx->createFunctionContext(func);
    std::vector<Fragment::edge> edges;
    edges.push_back({bb, A});
    edges.push_back({A, C});
    edges.push_back({C, D});
    edges.push_back({D, E});
    edges.push_back({C, E});
    edges.push_back({E, Fragment::EXIT});
    Fragment frag(*fctx.get(), bb, Fragment::EXIT, edges);

    Instruction* ret = &*E->begin();
    EXPECT_TRUE(frag.reachable(ret, ret));
    EXPECT_TRUE(frag.reachable(&*bb->begin(), ret));
    EXPECT_TRUE(frag.reachable(&*C->begin(), ret));
    EXPECT_TRUE(frag.reachable(&*D->begin(), ret));
    EXPECT_TRUE(frag.reachable(&*E->begin(), ret));
    EXPECT_TRUE(frag.reachable(&*C->begin(), &*E->begin()));

    EXPECT_FALSE(frag.reachable(ret, &*bb->begin()));
    EXPECT_FALSE(frag.reachable(ret, &*C->begin()));
    EXPECT_FALSE(frag.reachable(&*E->begin(), &*C->begin()));
}
