#include "./common.h"

#include "ModuleContext.h"
#include "FunctionContext.h"
#include "AbstractValue.h"
#include "FragmentDecomposition.h"
#include "ValueMapping.h"
#include "ConcreteState.h"
#include "ParamStrategy.h"

#include "domains/SimpleConstProp.h"
#include "domains/NumRels.h"
#include "domains/BitMask.h"
#include "domains/Product.h"
#include "domains/Predicates.h"
#include "domains/MemRegions.h"
#include "domains/Intervals.h"

namespace sprattus
{
namespace testing
{
const char* source = "define i32 @test(i32 %arg) #0 {\n"
                     "entry:\n"
                     "  %x = add i32 %arg, %arg\n"
                     "  %y = sub i32 %x, %arg\n"
                     "  ret i32 %x\n"
                     "}\n";

class SerializationTests : public ::testing::Test
{
  protected:
    std::string db_file;
    std::unique_ptr<llvm::Module> module;
    std::unique_ptr<const ModuleContext> mctx;
    llvm::Function* function;
    std::unique_ptr<FunctionContext> fctx;
    unique_ptr<ResultStore> store_ptr;
    ResultStore& store;
    RepresentedValue x_val;
    RepresentedValue y_val;
    RepresentedValue arg_val;
    FragmentDecomposition fdecomp;
    const Fragment* fragment;
    DB* dbp;

    SerializationTests()
        : db_file(std::tmpnam(nullptr)), module(parseAssembly(source)),
          mctx(new ModuleContext(module.get(), configparser::Config())),
          function(&*module->begin()),
          fctx(mctx->createFunctionContext(function)),
          store_ptr(new ResultStore(db_file)), store(*store_ptr),
          fdecomp(
              FragmentDecomposition::For((*fctx.get()), FragmentDecomposition::Edges)),
          dbp(store.DBP_)
    {
        fctx->setMemoryModel(std::unique_ptr<memory::BlockModel>(new memory::BlockModel(*fctx.get())));
        auto& symtbl = function->getValueSymbolTable();
        x_val = *(*fctx.get()).findRepresentedValue(symtbl.lookup("x"));
        y_val = *(*fctx.get()).findRepresentedValue(symtbl.lookup("y"));
        arg_val = *(*fctx.get()).findRepresentedValue(symtbl.lookup("arg"));
        assert(x_val != nullptr && arg_val != nullptr);

        for (auto& frag : fdecomp) {
            if (frag.getEnd() == Fragment::EXIT)
                fragment = &frag;
        }
        assert(fragment != nullptr);
    }

    ConcreteState makeConcreteState()
    {
        // get some (arbitrary) model and concrete value
        z3::solver solver((*fctx.get()).getZ3());
        solver.add((*fctx.get()).formulaFor(*fragment));
        solver.check();
        return ConcreteState(ValueMapping::atEnd((*fctx.get()), *fragment),
                             solver.get_model());
    }

    void checkSerialization(const AbstractValue& avalue)
    {
        std::ostringstream out;
        store.serialize(avalue, out);
        std::istringstream in(out.str());
        unique_ptr<AbstractValue> copy = store.deserialize(in, *fctx.get());

        EXPECT_TRUE(avalue.isJoinableWith(*copy));
        EXPECT_TRUE(copy->isJoinableWith(avalue));
        EXPECT_EQ(avalue, *copy);

        std::ostringstream os1, os2;
        os1 << avalue;
        os2 << *copy;
        EXPECT_EQ(os1.str(), os2.str());
    }

    ~SerializationTests()
    {
        // ensure that database connection is closed before deleting file
        store_ptr.reset();

        std::remove(db_file.c_str());
    }
};

TEST_F(SerializationTests, SimpleConstProp)
{
    domains::SimpleConstProp avalue((*fctx.get()), x_val);
    checkSerialization(avalue);

    avalue.havoc();
    checkSerialization(avalue);

    avalue.resetToBottom();
    avalue.updateWith(makeConcreteState());
    assert(!avalue.isBottom());
    checkSerialization(avalue);
}

TEST_F(SerializationTests, NumRels)
{
    auto avalue =
        domains::NumRels((*fctx.get()), *(*fctx.get()).findRepresentedValue(x_val),
                                  *(*fctx.get()).findRepresentedValue(arg_val));
    checkSerialization(avalue);
    avalue.updateWith(makeConcreteState());
    assert(!avalue.isBottom());
    checkSerialization(avalue);
    avalue.havoc();
    checkSerialization(avalue);
}

TEST_F(SerializationTests, Predicates)
{
    auto avalue =
        domains::Predicates((*fctx.get()), Expression(*(*fctx.get()).findRepresentedValue(x_val)).equals(*(*fctx.get()).findRepresentedValue(arg_val)));
    checkSerialization(avalue);
    avalue.updateWith(makeConcreteState());
    assert(!avalue.isBottom());
    checkSerialization(avalue);
    avalue.havoc();
    checkSerialization(avalue);
}

namespace // anonymous
{
domains::Predicates::pred_t test_pred = (domains::Predicates::pred_t)(
    [](Expression left, Expression right) { return left.equals(right); });
} // anonymous

TEST_F(SerializationTests, PredicatesWrapper)
{
    auto avalue =
        domains::PredicatesWrapper<&test_pred>((*fctx.get()), *(*fctx.get()).findRepresentedValue(x_val), *(*fctx.get()).findRepresentedValue(arg_val));
    checkSerialization(avalue);
    avalue.updateWith(makeConcreteState());
    assert(!avalue.isBottom());
    checkSerialization(avalue);
    avalue.havoc();
    checkSerialization(avalue);
}

TEST_F(SerializationTests, BitMask)
{
    domains::BitMask one_val((*fctx.get()), x_val);
    domains::BitMask two_val((*fctx.get()), x_val, arg_val);

    checkSerialization(one_val);
    checkSerialization(two_val);

    auto state = makeConcreteState();
    one_val.updateWith(state);
    two_val.updateWith(state);
    assert(!one_val.isTop() && !two_val.isTop());

    checkSerialization(one_val);
    checkSerialization(two_val);

    one_val.havoc();
    two_val.havoc();

    checkSerialization(one_val);
    checkSerialization(two_val);
}

TEST_F(SerializationTests, Product)
{
    using namespace domains;

    llvm::BasicBlock* bb = &*function->begin();
    Product avalue((*fctx.get()));
    avalue.add(params::ForValues<SimpleConstProp>((*fctx.get()), bb, true));
    avalue.add(params::ForValuePairs<BitMask>((*fctx.get()), bb, true));
    avalue.finalize();

    checkSerialization(avalue);
    avalue.updateWith(makeConcreteState());
    checkSerialization(avalue);
    avalue.havoc();
    checkSerialization(avalue);
}

TEST_F(SerializationTests, Interval)
{
    domains::Interval avalue((*fctx.get()),
            *(*fctx.get()).findRepresentedValue(x_val));
    checkSerialization(avalue);
}

TEST_F(SerializationTests, ValidRegion)
{
    domains::ValidRegion avalue((*fctx.get()),
            *(*fctx.get()).findRepresentedValue(x_val));
    checkSerialization(avalue);
}

TEST_F(SerializationTests, ConstantRegion)
{
    domains::ConstantRegion avalue((*fctx.get()),
            *(*fctx.get()).findRepresentedValue(x_val));
    checkSerialization(avalue);
}

TEST_F(SerializationTests, NoAlias)
{
    domains::NoAlias avalue((*fctx.get()),
            *(*fctx.get()).findRepresentedValue(x_val),
            *(*fctx.get()).findRepresentedValue(arg_val));
    checkSerialization(avalue);
}

TEST_F(SerializationTests, VariableRegion)
{
    domains::VariableRegion avalue((*fctx.get()),
            *(*fctx.get()).findRepresentedValue(x_val),
            *(*fctx.get()).findRepresentedValue(y_val),
            *(*fctx.get()).findRepresentedValue(arg_val));
    checkSerialization(avalue);
}

TEST_F(SerializationTests, RestrictedVarRegion)
{
    domains::RestrictedVarRegion avalue((*fctx.get()),
            *(*fctx.get()).findRepresentedValue(x_val),
            *(*fctx.get()).findRepresentedValue(y_val),
            *(*fctx.get()).findRepresentedValue(arg_val));
    checkSerialization(avalue);
}

TEST_F(SerializationTests, DatabaseWriteRead)
{
    EXPECT_NE(nullptr, dbp);

    DBT key, data;
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    int key_val = 12345;
    key.data = &key_val;
    key.size = sizeof(key_val);

    char data_val[] = "Hello";
    data.data = &data_val;
    data.size = sizeof(data_val);

    int ret = dbp->put(dbp, nullptr, &key, &data, 0);
    EXPECT_EQ(0, ret);

    memset(data_val, '\0', sizeof(data_val));
    data.ulen = sizeof(data_val);
    data.flags = DB_DBT_USERMEM;

    EXPECT_NE("Hello", std::string(data_val));
    ret = dbp->get(dbp, nullptr, &key, &data, 0);
    EXPECT_EQ(0, ret);
    EXPECT_EQ("Hello", std::string(data_val));
}

TEST_F(SerializationTests, ResultStoreWriteRead)
{
    domains::SimpleConstProp avalue((*fctx.get()), x_val);
    avalue.updateWith(makeConcreteState());
    assert(!avalue.isBottom());

    store.put(ResultStore::Key(0), avalue);
    unique_ptr<AbstractValue> avalue_copy =
        store.get(ResultStore::Key(0), *fctx.get());
    EXPECT_EQ(avalue, *avalue_copy);

    avalue_copy = store.get(ResultStore::Key(1), *fctx.get());
    EXPECT_EQ(nullptr, avalue_copy);
}
} // namespace testing
} // namespace sprattus
