#include <memory>
#include <iterator>

#include "./common.h"

#include "ModuleContext.h"
#include "FunctionContext.h"
#include "Fragment.h"
#include "ValueMapping.h"

using namespace sprattus;
using namespace llvm;

TEST(ValueMappingTest, SelfLoop)
{
    const char* src = "define i32 @test(i32 %a) #0 {\n"
                      "bb:\n"
                      "  br label %head\n"
                      "\n"
                      "head:\n"
                      "  %i = phi i32 [ 0, %bb ], [ %i_next, %head ]\n"
                      "  %test_i = icmp slt i32 %i, %a\n"
                      "  %i_next = add nsw i32 %i, 1\n"
                      "  br i1 %test_i, label %head, label %end\n"
                      "\n"
                      "end:\n"
                      "  ret i32 %i\n"
                      "}\n";

    auto module = parseAssembly(src);
    Function* func = &*module->begin();
    BasicBlock* bb, *head, *end;
    unpack(std::tie(bb, head, end), pointerVector(func->getBasicBlockList()));

    auto mctx = make_unique<const ModuleContext>(module.get(),
                                                 configparser::Config());
    auto fctx = mctx->createFunctionContext(func);
    std::vector<Fragment::edge> edges = {Fragment::edge(head, head)};
    Fragment frag(*fctx.get(), head, head, edges);

    Instruction* i, *test_i, *i_next, *br;
    unpack(std::tie(i, test_i, i_next, br), pointerVector(head->getInstList()));

    ValueMapping vm_start = ValueMapping::atBeginning((*fctx.get()), frag);
    ValueMapping vm_test_i_before = ValueMapping::before((*fctx.get()),
                                                         frag, test_i);
    EXPECT_EQ("test_i_0", repr(vm_start[test_i]));
    EXPECT_EQ("test_i_0", repr(vm_test_i_before[test_i]));
    EXPECT_EQ("i_next_0", repr(vm_start[i_next]));
    EXPECT_EQ("i_next_0", repr(vm_test_i_before[i_next]));
    EXPECT_EQ("i_0", repr(vm_start[i]));
    EXPECT_EQ("i_0", repr(vm_test_i_before[i]));

    ValueMapping vm_end = ValueMapping::atEnd((*fctx.get()), frag);
    ValueMapping vm_i_after = ValueMapping::after((*fctx.get()), frag, i);
    EXPECT_EQ("i_1", repr(vm_end[i]));
    EXPECT_EQ("i_1", repr(vm_i_after[i]));
    EXPECT_EQ("test_i_1", repr(vm_end[test_i]));
    EXPECT_EQ("test_i_1", repr(vm_i_after[test_i]));
    EXPECT_EQ("i_next_1", repr(vm_end[i_next]));
    EXPECT_EQ("i_next_1", repr(vm_i_after[i_next]));

    ValueMapping vm_i_before = ValueMapping::before((*fctx.get()), frag, i);
    EXPECT_EQ("i_0", repr(vm_i_before[i]));
    EXPECT_EQ("test_i_1", repr(vm_i_before[test_i]));
    EXPECT_EQ("i_next_1", repr(vm_i_before[i_next]));

    ValueMapping vm_br = ValueMapping::before((*fctx.get()), frag, br);
    EXPECT_EQ("test_i_1", repr(vm_br[test_i]));
    EXPECT_EQ("i_next_1", repr(vm_br[i_next]));
    EXPECT_EQ("i_0", repr(vm_br[i]));
}

TEST(ValueMappingTest, Memory)
{
    const char* src = "define i32 @test(i32* %ptr) #0 {\n"
                      "start:\n"
                      "    br label %bb\n"
                      "bb:\n"
                      "    %t1 = load i32* %ptr\n"
                      "    %t2 = add i32 %t1, 1\n"
                      "    store i32 %t2, i32* %ptr\n"
                      "    %t3 = load i32* %ptr\n"
                      "    %t4 = add i32 %t2, %t3\n"
                      "    br label %end\n"
                      "end:\n"
                      "    ret i32 %t4\n"
                      "}\n";

    auto module = parseAssembly(src);
    Function* func = &*module->begin();
    BasicBlock* start, *bb, *end;
    unpack(std::tie(start, bb, end), pointerVector(func->getBasicBlockList()));

    auto mctx = make_unique<const ModuleContext>(module.get(),
                                                 configparser::Config());
    auto fctx = mctx->createFunctionContext(func);
    std::vector<Fragment::edge> edges = {Fragment::edge(start, bb),
                                         Fragment::edge(bb, end)};
    Fragment frag((*fctx.get()), start, end, edges);

    Instruction* t1, *t2, *store, *t3, *t4, *br;
    unpack(std::tie(t1, t2, store, t3, t4, br),
           pointerVector(bb->getInstList()));

    EXPECT_EQ("mem_bb_0",
              repr(ValueMapping::atLocation((*fctx.get()), frag, bb).memory()));
    EXPECT_EQ("mem_bb_0",
              repr(ValueMapping::before((*fctx.get()), frag, t1).memory()));
    EXPECT_EQ("mem_bb_0",
              repr(ValueMapping::before((*fctx.get()), frag, t2).memory()));
    EXPECT_EQ("mem_bb_0",
              repr(ValueMapping::before((*fctx.get()), frag, store).memory()));
    EXPECT_EQ("mem_bb_1",
              repr(ValueMapping::before((*fctx.get()), frag, t3).memory()));
    EXPECT_EQ("mem_bb_1",
              repr(ValueMapping::before((*fctx.get()), frag, t4).memory()));
    EXPECT_EQ("mem_bb_1",
              repr(ValueMapping::before((*fctx.get()), frag, br).memory()));

    EXPECT_EQ("mem_end_0",
              repr(ValueMapping::atLocation((*fctx.get()), frag, end).memory()));
}

TEST(ValueMappingTest, SelfLoopMemory)
{
    const char* src = "define i32 @test(i32 %a, i32* %ptr) #0 {\n"
                      "bb:\n"
                      "  br label %head\n"
                      "\n"
                      "head:\n"
                      "  %i = phi i32 [ 0, %bb ], [ %i_next, %head ]\n"
                      "  store i32 %i, i32* %ptr\n"
                      "  %test_i = icmp slt i32 %i, %a\n"
                      "  %i_next = add nsw i32 %i, 1\n"
                      "  br i1 %test_i, label %head, label %end\n"
                      "\n"
                      "end:\n"
                      "  ret i32 %i\n"
                      "}\n";

    auto module = parseAssembly(src);
    Function* func = &*module->begin();
    BasicBlock* bb, *head, *end;
    unpack(std::tie(bb, head, end), pointerVector(func->getBasicBlockList()));

    auto mctx = make_unique<const ModuleContext>(module.get(),
                                                 configparser::Config());
    auto fctx = mctx->createFunctionContext(func);
    std::vector<Fragment::edge> edges = {Fragment::edge(head, head)};
    Fragment frag(*fctx.get(), head, head, edges);

    Instruction* i, *store, *test_i, *i_next, *br;
    unpack(std::tie(i, store, test_i, i_next, br),
           pointerVector(head->getInstList()));

    auto mvar_before = [&fctx, &frag](Instruction* inst) {
        return repr(ValueMapping::before(*fctx.get(), frag, inst).memory());
    };

    EXPECT_EQ("mem_head_0", mvar_before(i));
    EXPECT_EQ("mem_head_0", mvar_before(store));
    EXPECT_EQ("mem_head_1", mvar_before(test_i));
    EXPECT_EQ("mem_head_1", mvar_before(i_next));
    EXPECT_EQ("mem_head_1", mvar_before(br));
}
